/*
  Copyright (C) 2005-2019 Steven L. Scott

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License along
  with this library; if not, write to the Free Software Foundation, Inc., 51
  Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
*/

#include "Models/Glm/PosteriorSamplers/OrdinalLogitImputer.hpp"
#include "distributions.hpp"

namespace BOOM {

  double OrdinalLogitImputer::impute(
      RNG &rng, double eta, double lower_cutpoint, double upper_cutpoint) {
    return eta + qlogis(runif_mt(
        rng, plogis(lower_cutpoint - eta), plogis(upper_cutpoint - eta)));
  }
  
}

