// Copyright 2018 Google LLC. All Rights Reserved.
/*
  Copyright (C) 2008 Steven L. Scott

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
*/
#ifndef BOOM_OUTPUT_TABLE_HPP
#define BOOM_OUTPUT_TABLE_HPP
#include "cpputil/string_utils.hpp"
#include "uint.hpp"

namespace BOOM {

  class OutputTable {
   public:
    explicit OutputTable(uint pad = 2);
    std::vector<std::string> &column(uint i);
    OutputTable &add_row(const std::vector<std::string> &);
    OutputTable &add_column(const std::vector<std::string> &);
    OutputTable &add_to_column(const std::string &, uint i);

    std::ostream &print(std::ostream &) const;
    void equalize_rows();

   private:
    uint pad_;
    std::vector<std::vector<std::string>> cols_;
  };

}  // namespace BOOM
#endif  // BOOM_OUTPUT_TABLE_HPP
